####
# bge_game3_core: Full python game structure for the Blender Game Engine
# Copyright (C) 2019  DaedalusMDW @github.com (Daedalus_MDW @blenderartists.org)
# https://github.com/DaedalusMDW/bge_game3_core
#
# This file is part of bge_game3_core.
#
#    bge_game3_core is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    bge_game3_core is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with bge_game3_core.  If not, see <http://www.gnu.org/licenses/>.
#
####

## PLAYER CLASS ##

import math

from bge import logic, events, render
from mathutils import *



PLAYER = None


def main(cont):
	global PLAYER

	owner = cont.owner
	scene = owner.scene

	if PLAYER == None:
		PLAYER = CorePlayer()

	PLAYER.RUN()


class CorePlayer:

	NAME = "Player"

	SPEED = 0.1
	JUMP = 6
	ACCEL = 20
	ACCEL_FAST = 10
	ACCEL_STOP = 10
	SLOPE = 60
	MOVERUN = True
	SIDESTEP = False
	AIR_DRAG = (0.67, 0.67, 0.33)
	GRAV_DAMAGE = 10

	INTERACT = 2
	EYE_H = 1.6
	GND_H = 1.0
	EDGE_H = 2.0
	WALL_DIST = 0.4
	OFFSET = (0, 0.0, 0.2)
	SLOPE_SPEED = 1.0
	SLOPE_BIAS = 0.0
	CROUCH_H = 0.4
	CROUCH_SCALE = 0.2 #Capsule is 50% effected by scale

	CAM_TYPE = "FIRST"
	CAM_FOV = 90
	CAM_SLOW = 10
	CAM_RANGE = (0.75, 6) #0.75 6
	CAM_HEIGHT = 0.15
	CAM_STEPS = 7
	CAM_ZOOM = 2
	CAM_MIN = 0.2

	def __init__(self):
		scene = logic.getCurrentScene()
		cont = logic.getCurrentController()
		owner = cont.owner

		chr = owner.childrenRecursive
		self.objects = {
			"Root":owner,
			"Character":chr["Actor"],
			"Rig":chr["Actor.Rig"],
			"Mesh":chr["Actor.Mesh"],
			"Viewport":chr["Viewport"],
			"CamOrb":chr["Viewport.Root"],
			"CamRot":chr["Viewport.Rotate"],
			"Camera":chr["Viewport.Camera"]
			}

		self.ANIMOBJ = self.objects.get("Rig", None)

		self.defaultStates()

		self.controller = cont
		self.jump_state = "NONE"
		self.jump_timer = 0
		self.lastaction = [None,0]

		self.resetGroundRay()
		self.resetAcceleration()

		self.motion = {
			"Move": Vector([0,0]),
			"Rotate": Vector([0,0,0]),
			"Climb": 0,
			"Local": Vector([0,0,0]),
			"World": Vector([0,0,0])
			}

		self.data = {"HEALTH":100, "ENERGY":100, "RECHARGE":0.02, "SPEED":self.SPEED,
			"JUMP":self.JUMP, "RUN":self.MOVERUN, "STRAFE":self.SIDESTEP}

		self.camdata = {
			"State": self.CAM_TYPE,
			"Zoom": self.CAM_ZOOM,
			"Distance": None,
			"FOV": self.CAM_FOV,
			"Slow": self.CAM_SLOW,
			"POS": [0,0,0],
			"ZR": [0,0,0],
			"XR": 0}

		self.data["PHYSICS"] = "DYNAMIC"
		self.data["CROUCH"] = 0

		self.gravity = Vector(scene.gravity)

		owner.setDamping(0, 0)

		self.doPlayerAnim("RESET")

	def defaultStates(self):
		self.active_pre = [self.PR_LastVelocity]
		self.active_state = self.ST_Walking
		self.active_post = [self.PS_SetVisible, self.PS_GroundTrack, self.PS_Camera]

	def toDeg(self, rad):
		return math.degrees(rad)

	def getLocalSpace(self, obj, pnt):
		pnt = Vector(pnt)
		lp = pnt - obj.worldPosition
		lp = obj.worldOrientation.inverted()*lp
		return lp

	def getWorldSpace(self, obj, pnt):
		pnt = Vector(pnt)
		wp = obj.worldOrientation*pnt
		wp = wp + obj.worldPosition
		return wp

	def resetGroundRay(self):
		self.groundhit = None
		self.groundold = None
		self.groundobj = None
		self.groundchk = False
		self.groundvel = Vector([0,0,0])
		self.groundpos = [Vector([0,0,0]), Vector([0,0,0])]
		self.groundori = [Matrix.Identity(3), Matrix.Identity(3)]

		self.gndraybias = self.GND_H

	def resetAcceleration(self, speed=(0,0,0)):
		self.accel_start = Vector([0,0,0])
		self.accel_move = Vector(speed)
		self.accel_end = Vector(speed)
		self.accel_timer = 0
		self.accel_stand = -1

	def alignCamera(self, factor=1.0, axis=(0,1,0), up=None):
		vref = self.objects["Root"].getAxisVect(axis)
		if up == None:
			up = self.objects["Root"].getAxisVect((0,0,1))
		self.objects["CamOrb"].alignAxisToVect(vref, 1, factor)
		self.objects["CamOrb"].alignAxisToVect(up, 2, 1.0)

	def alignPlayer(self, factor=1.0, axis=None, up=None):
		if axis == None:
			axis = self.objects["CamOrb"].getAxisVect([0,1,0])
		self.objects["Root"].alignAxisToVect(axis, 1, factor)
		if up == None:
			self.alignToGravity()
		elif up != False:
			self.objects["Root"].alignAxisToVect(up, 2, 1.0)

	def alignToGravity(self, obj=None, axis=2, neg=False):
		if obj == None:
			obj = self.objects["Root"]
		if self.gravity.length >= 0.1:
			grav = self.gravity.normalized()
			if neg == False:
				grav = -grav
			obj.alignAxisToVect(grav, axis, 1.0)

	def doAnim(self, OBJECT=None, NAME="NONE", FRAME=(0, 60), LAYER=0, PRIORITY=0, MODE="PLAY", BLEND=0, KEY=False, STOP=False, CHECK=False, SET=None):

		if type(OBJECT) is str:
			if OBJECT not in self.objects:
				OBJECT = None
			else:
				OBJECT = self.objects[OBJECT]

		if OBJECT == None:
			if self.ANIMOBJ != None:
				OBJECT = self.ANIMOBJ
			elif self.objects.get("Root", None) != None:
				OBJECT = self.objects["Root"]
			else:
				return

		if SET != None:
			OBJECT.setActionFrame(SET, LAYER)
			return

		if CHECK == "FRAME":
			ACTCHK = OBJECT.getActionFrame(LAYER)
			return ACTCHK
		if CHECK == True:
			ACTCHK = OBJECT.isPlayingAction(LAYER)
			return ACTCHK

		if STOP == True:
			OBJECT.stopAction(LAYER)
			return

		if NAME == "NONE":
			NAME = OBJECT.name
		if NAME == None:
			return

		SPEED = 1
		LYRWGHT = 0
		IPO_FLG = 0

		if MODE == "PLAY":
			PLAYTYPE = logic.KX_ACTION_MODE_PLAY

		elif MODE == "LOOP":
			PLAYTYPE = logic.KX_ACTION_MODE_LOOP

		BLENDTYPE = logic.KX_ACTION_BLEND_BLEND

		OBJECT.playAction(NAME, FRAME[0], FRAME[1], LAYER, PRIORITY, BLEND, PLAYTYPE, LYRWGHT, IPO_FLG, SPEED, BLENDTYPE)
		if KEY == True:
			OBJECT.reinstancePhysicsMesh()

	def getInputs(self):
		owner = self.objects["Root"]
		sensor = self.controller.sensors

		kbe = logic.keyboard.events
		mse = logic.mouse.events

		self.motion["Move"][0] = sensor["STRAFE_RIGHT"].positive - sensor["STRAFE_LEFT"].positive
		self.motion["Move"][1] = sensor["FORWARD"].positive - sensor["BACKWARD"].positive
		self.motion["Climb"] = sensor["JUMP"].positive - sensor["CROUCH"].positive

		self.motion["Rotate"][0] = 0
		self.motion["Rotate"][1] = 0
		self.motion["Rotate"][2] = 0

		self.data["RUN"] = owner.get("PLR_RUN", True)

		self.camdata["Zoom"] += (mse[events.WHEELDOWNMOUSE] in [1,2]) - (mse[events.WHEELUPMOUSE] in [1,2])

		newcam = owner.get("CAMERA", "THIRD")
		if self.camdata["State"] != newcam:
			self.setCameraState(newcam)

	def setCameraState(self, state=None, load=False):
		if state == None:
			state = self.camdata["State"]
		else:
			self.camdata["State"] = state

		if state == "SHOULDER":
			pos = [self.CAM_SHSIDE, 0, self.EYE_H-self.GND_H-self.CAM_MIN]
		else:
			pos = [0, 0, self.EYE_H-self.GND_H]

		self.camdata["FOV"] = self.CAM_FOV
		self.camdata["State"] = state

	def setPhysicsType(self, mode=None):
		owner = self.objects["Root"]

		if mode == None:
			mode = self.data["PHYSICS"]

		if mode == "NONE":
			owner.disableRigidBody()
			owner.suspendDynamics()
		else:
			owner.restoreDynamics()
			if mode == "RIGID":
				owner.enableRigidBody()
			else:
				owner.disableRigidBody()

		self.data["PHYSICS"] = mode

	def getGroundPoint(self, obj):
		if self.groundchk == True:
			print("WARNING: Ground reference already updated...")
			return

		newpos = obj.worldPosition.copy()
		newori = obj.worldOrientation.copy()

		self.groundpos.insert(0, newpos)
		self.groundpos.pop()
		self.groundori.insert(0, newori)
		self.groundori.pop()

		self.groundchk = True

		if self.groundobj != obj:
			self.groundobj = obj
			self.groundvel *= 0
			self.groundpos[1] = self.groundpos[0].copy()
			self.groundori[1] = self.groundori[0].copy()

	def checkWall(self, z=True, axis=None, simple=None, prop="GROUND"):
		owner = self.objects["Root"]

		if axis == None:
			axis = owner.getAxisVect((0,1,0))
		else:
			axis = Vector(axis)
		dist = self.WALL_DIST+0.3
		angle = -1

		if simple != None:
			dist = simple

		WALLOBJ, WALLPNT, WALLNRM = owner.rayCast(owner.worldPosition+axis, None, dist, "", 1, 0, 0)

		if WALLOBJ != None:
			if simple != None:
				return WALLOBJ, WALLPNT, WALLNRM
			if z == False:
				WALLNRM = owner.localOrientation.inverted()*WALLNRM
				WALLNRM[2] = 0
				WALLNRM.normalize()
				WALLNRM = owner.localOrientation*WALLNRM
			if WALLNRM.length > 0.01:
				angle = axis.angle(WALLNRM, 0)
				angle = round(self.toDeg(angle), 2)

		else:
			if simple != None:
				return None

		return angle, WALLNRM

	def checkGround(self, simple=False, ray=None):
		owner = self.objects["Root"]
		ground = None
		angle = 0
		move = self.motion["World"].copy()
		gndto = owner.getAxisVect((0,0,-1))+owner.worldPosition
		gndbias = 0
		tan = 0

		if ray == None:
			raylist = (gndto, None, self.gndraybias)
		else:
			raylist = ray

		rayOBJ, rayPNT, rayNRM = owner.rayCast(raylist[0], raylist[1], raylist[2]+0.3, "", 1, 1, 0)

		if rayOBJ == None or self.gravity.length < 0.1:
			self.gndraybias = self.GND_H

			if simple == True:
				return None

		else:
			ground = [rayOBJ, rayPNT, rayNRM]

			angle = owner.getAxisVect((0,0,1)).angle(rayNRM, 0)
			angle = round(self.toDeg(angle), 2)
			gndbias = (angle/90)*self.SLOPE_BIAS

			if self.jump_state not in ["NONE", "CROUCH"] and ray == None:
				dist = owner.worldPosition-rayPNT
				if dist.length > self.GND_H+gndbias:
					ground = None
				else:
					self.gndraybias = self.GND_H+gndbias

				#if angle > self.SLOPE:
				#	ground = None

			if simple == True:
				self.gndraybias = self.GND_H+gndbias
				return ground

		if ground != None:
			self.getGroundPoint(rayOBJ)

			if self.groundold != None:
				gnddiff = self.groundold[1]-rayPNT
				gnddiff = owner.worldOrientation.inverted()*gnddiff
				if abs(gnddiff[2]) < 0.4:
					self.gndraybias += gnddiff[2]

			self.groundhit = ground
			self.gndraybias += ((self.GND_H+gndbias)-self.gndraybias)*0.2

		return ground, angle

	def doJump(self, height=None, move=0.5, align=False):
		owner = self.objects["Root"]

		self.jump_state = "JUMP"
		self.jump_timer = 0

		owner.setDamping(0, 0)

		if height == None:
			height = self.data["JUMP"]

		align = owner.localLinearVelocity.copy()*align
		align[2] = 0
		if align.length > 0.02 and move > 0:
			if self.data["STRAFE"] == False and self.camdata["State"] == "THIRD":
				self.alignPlayer(axis=owner.getAxisVect(align))

		owner.localLinearVelocity[0] *= move
		owner.localLinearVelocity[1] *= move

		owner.localLinearVelocity[2] = height

		self.doPlayerAnim("JUMP", 5)

	def doMovement(self, vec, mx=0):
		owner = self.objects["Root"]

		world = self.motion["World"].copy()
		local = self.motion["Local"].copy()

		if self.groundhit != None:
			rayNRM = self.groundhit[2]
		else:
			rayNRM = owner.getAxisVect((0,0,1))

		angle = owner.getAxisVect((0,0,1)).angle(rayNRM, 0)
		angle = round(self.toDeg(angle), 2)

		dot = 0
		if angle > 5 and local.length > 0.01:
			slvec = owner.worldOrientation.inverted()*rayNRM
			slvec[2] = 0

			slang = slvec.normalized().angle(local.normalized(), 0)
			slang = round(self.toDeg(slang), 2)

			if angle <= self.SLOPE:
				dot = abs(1-(slang/90))*self.SLOPE_SPEED

		slope = 1-(dot*(angle/90))

		vref = self.objects["CamOrb"].getAxisVect(vec)
		mref = (vref*mx*slope)

		if self.ACCEL < 1:
			move = mref.copy()
		else:
			check = self.accel_end-mref
			if check.length > 0.001:
				self.accel_start = self.accel_move.copy()
				self.accel_end = mref.copy()
				self.accel_timer = 0
				if self.accel_stand == -1 and self.accel_start.length < 0.001:
					self.accel_stand = 0
				elif self.accel_start.normalized().dot(self.accel_end.normalized()) < 0:
					self.accel_stand = 0

			rate = (1/self.ACCEL)
			if self.accel_stand == -1:
				rate = (1/self.ACCEL_FAST)
			if self.accel_end.length < 0.001:
				rate = (1/self.ACCEL_STOP)

			offset = mref-self.accel_start
			self.accel_move += offset*rate
			check = mref-self.accel_move

			greater = offset.normalized().dot(check.normalized())

			if greater < 0 or check.length < 0.001:
				self.accel_move = mref.copy()

			if self.accel_timer < self.ACCEL:
				self.accel_timer += 1
			else:
				self.accel_timer = self.ACCEL

			if self.accel_stand >= 0:
				self.accel_stand += 1
			if self.accel_stand >= self.ACCEL:
				self.accel_stand = -1

			move = self.accel_move.copy()

		if angle > 5 and angle <= self.SLOPE and world.length > 0.001:
			tango = rayNRM.angle(world.normalized(), 0)
			tango = round(self.toDeg(tango), 2)

			tan = math.tan(math.radians(tango-90))*(world.length/60)
			tangle = owner.getAxisVect((0,0,1))*tan
			move += tangle

		owner.worldLinearVelocity = move*60

	def doMoveAlign(self, axis=None, up=None, margin=0.002):
		owner = self.objects["Root"]

		if axis != None:
			align = Vector(axis)
		elif self.data["STRAFE"] == True or self.camdata["State"] != "THIRD":
			align = self.objects["CamOrb"].getAxisVect([0,1,0])
		else:
			align = owner.worldLinearVelocity*(1/60)

		align = owner.worldOrientation.inverted()*align
		align[2] = 0
		align = owner.worldOrientation*align

		if align.length >= margin:
			align.normalize()
			dref = owner.getAxisVect((0,1,0))
			dot = 1-((dref.dot(align)*0.5)+0.5)
			dot = 0.2+((dot**2)*0.5)

			self.alignPlayer(dot, align, up)

		elif up == None:
			self.alignToGravity()
		elif up != False:
			owner.alignAxisToVect(up, 2, 1.0)

	def doPlayerOnGround(self):
		owner = self.objects["Root"]
		scene = owner.scene

		move = self.motion["Move"].normalized()

		if self.data["RUN"] == False or self.motion["Move"].length <= 0.7 or self.data["SPEED"] <= 0:
			mx = 0.035
		else:
			mx = self.data["SPEED"]

		self.doMovement((move[0], move[1], 0), mx)
		self.doMoveAlign(up=False)
		self.doPlayerAnim("MOVE", blend=10)

		## Jump/Crouch ##
		sen = self.controller.sensors
		if sen["CROUCH"].status == 2:
			self.doCrouch(True)

		elif sen["JUMP"].status == 1:
			if self.jump_timer == 1:
				self.jump_timer = 2
		elif sen["JUMP"].status == 2:
			if self.jump_timer >= 2:
				self.jump_timer += 1
			if self.jump_timer >= 7:
				self.doJump(move=0.8, align=True)
		elif sen["JUMP"].status == 3:
			if self.jump_timer >= 1 and self.jump_timer < 7:
				self.doJump(height=self.JUMP/1.5, move=1.0, align=False)
		else:
			self.jump_timer = 1

	def doPlayerAnim(self, action="MOVE", blend=10):
		if self.ANIMOBJ == None:
			return

		if action == "JUMP":
			self.doAnim(NAME="Jumping", FRAME=(0,20), PRIORITY=2, MODE="PLAY", BLEND=10)
			self.lastaction = [action, 0]
			return

		if action == "FALLING":
			self.doAnim(NAME="Jumping", FRAME=(40,40), PRIORITY=3, MODE="LOOP", BLEND=blend)
			self.lastaction = [action, 0]
			return

		if action in ["IDLE", "RESET"]:
			invck = 0

			if action == "RESET":
				self.doAnim(NAME="Jumping", FRAME=(0+invck,0+invck))
			else:
				self.doAnim(NAME="Jumping", FRAME=(0+invck,0+invck), PRIORITY=3, MODE="LOOP", BLEND=blend)
			self.lastaction = [action, 0]
			return

		linLV = self.motion["Local"]

		if action == "CROUCH":
			if linLV.length > 0.1:
				self.doAnim(NAME="Crouching", FRAME=(0,80), PRIORITY=3, MODE="LOOP", BLEND=blend)
			else:
				self.doAnim(NAME="Crouching", FRAME=(0,0), PRIORITY=3, MODE="LOOP", BLEND=blend)
			self.lastaction = [action, 0]
			return

		action = "IDLE"
		walk = False
		stair = ""
		setframe = 0

		if linLV.length > 0.1:
			if self.ACCEL > 10 and self.accel_stand >= 0:
				blend = self.ACCEL

			if linLV.length < (0.05*60):
				walk = True

			action = "FORWARD"

			if linLV[1] < 0:
				action = "BACKWARD"
			if linLV[0] > 0.5 and abs(linLV[1]) < 0.5:
				action = "STRAFE_R"
			if linLV[0] < -0.5 and abs(linLV[1]) < 0.5:
				action = "STRAFE_L"

			if self.groundhit != None and self.objects["Root"] != None:
				angle = self.objects["Root"].getAxisVect((0,0,1)).angle(self.groundhit[2], 0)
				angle = round(self.toDeg(angle), 2)

				if angle > 20 and angle <= self.SLOPE:
					stair = ".Stairs"

		## ANIMATIONS ##
		if action == "FORWARD":
			if walk == True:
				self.doAnim(NAME="Walking", FRAME=(0,59), PRIORITY=3, MODE="LOOP", BLEND=blend)
				setframe = 59
			else:
				self.doAnim(NAME="Running"+stair, FRAME=(0,39), PRIORITY=3, MODE="LOOP", BLEND=blend)
				setframe = 39

		elif action == "BACKWARD":
			#if walk == True:
			self.doAnim(NAME="Walking", FRAME=(59,0), PRIORITY=3, MODE="LOOP", BLEND=blend)
			setframe = 59
			#else:
			#	self.doAnim(NAME="Running"+stair, FRAME=(39,0), PRIORITY=3, MODE="LOOP", BLEND=blend)
			#	setframe = 39

		else:
			self.lastaction = ["IDLE", 0]
			self.doPlayerAnim("IDLE")
			return

		if self.lastaction[0] != action:
			self.lastaction[0] = action
			self.doAnim(SET=self.lastaction[1]*setframe)

		self.lastaction[1] = self.doAnim(CHECK="FRAME")/setframe

	def doCameraCollision(self):
		camera = self.objects["Camera"]
		rotate = self.objects["CamRot"]

		margin = 1
		height = self.CAM_HEIGHT
		minz = self.CAM_MIN

		if self.camdata["Zoom"] < 0:
			self.camdata["Zoom"] = 0

		if self.camdata["Zoom"] > self.CAM_STEPS:
			self.camdata["Zoom"] = self.CAM_STEPS

		steps = (self.CAM_RANGE[1]-self.CAM_RANGE[0])/self.CAM_STEPS
		dist = (steps*self.camdata["Zoom"])+self.CAM_RANGE[0]

		camLX = 0
		camLY = -dist
		camLZ = dist*height

		obj = self.objects["Root"]
		if obj != None:
			ray = dist+margin
			rayto = Vector([0, -ray, ray*height])
			rayto = self.getWorldSpace(rotate, rayto)

			hyp = ray**2 + (ray*height)**2

			rayOBJ, rayPNT, rayNRM = obj.rayCast(rayto, rotate, hyp**0.5, "CAMERA", 1, 1, 0)

			if rayOBJ:
				local = self.getLocalSpace(rotate, rayPNT)

				margin = margin*((local.length)/ray)

				camLX = 0
				camLY = local[1]+margin
				camLZ = local[2]-(margin*height)

		if camLY > 0:
			camLY = 0
		if camLZ < minz:
			camLZ = minz

		camera.localPosition[0] = camLX
		camera.localPosition[1] = camLY
		camera.localPosition[2] = camLZ

	def doCrouch(self, state):
		self.jump_timer = 0
		if state == True or self.data["CROUCH"] != 0:
			self.jump_state = "CROUCH"
			self.objects["Root"].localScale[2] = self.CROUCH_SCALE
			self.objects["Character"].localScale[2] = 1/self.CROUCH_SCALE
			self.active_state = self.ST_Crouch
		elif state == False:
			self.jump_state = "NONE"
			self.objects["Root"].localScale[2] = 1
			self.objects["Character"].localScale[2] = 1
			self.objects["Character"].localPosition = (0,0,0)
			self.active_state = self.ST_Walking

	## INIT STATE ##
	def ST_Startup(self):
		if self.jump_state == "FLYING" and self.objects["Root"] != None:
			self.ST_Advanced_Set()

	## PRE ##
	def PR_LastVelocity(self):
		owner = self.objects["Root"]

		gndvel = self.groundvel.copy()

		if owner != None:
			linWV = owner.worldLinearVelocity-gndvel
			linLV = owner.worldOrientation.inverted()*linWV
			linLV[2] = 0
			linWV = owner.worldOrientation*linLV
		else:
			linWV = Vector([0,0,0])
			linLV = Vector([0,0,0])

		self.motion["Local"] = linLV
		self.motion["World"] = linWV

	## POST ##
	def PS_GroundTrack(self):
		owner = self.objects["Root"]
		self.groundchk = False

		owner.setDamping(0, 0)

		if self.groundhit == None or self.jump_state not in ["NONE", "CROUCH", "JUMP"]:
			drag = self.AIR_DRAG
			dragX = owner.localLinearVelocity[0]*drag[0] #0.67
			dragY = owner.localLinearVelocity[1]*drag[1] #0.67
			dragZ = owner.localLinearVelocity[2]*drag[2] #0.33
			owner.applyForce((-dragX, -dragY, -dragZ), True)
			self.groundvel *= 0
			self.groundhit = None
			self.groundold = None
			self.groundobj = None
			return

		rayOBJ = self.groundobj
		wp = owner.worldPosition

		locOLD = wp - self.groundpos[1]
		posOLD = self.groundori[1].inverted()*locOLD

		locNEW = wp - self.groundpos[0]
		posNEW = self.groundori[0].inverted()*locNEW

		local = posOLD - posNEW
		offset = self.groundori[0]*local

		if self.data["PHYSICS"] == "NONE":
			owner.applyMovement(offset, False)
		else:
			owner.worldLinearVelocity += offset*60

		yvec = owner.getAxisVect((0,1,0))
		rotOLD = self.groundori[1].inverted()*yvec
		rotOLD = self.groundori[0]*rotOLD
		euler = yvec.rotation_difference(rotOLD).to_euler()

		owner.applyRotation((0,0,euler[2]), True)

		owner.applyForce(-self.gravity, False)

		self.groundvel = offset.copy()*60
		self.groundold = self.groundhit.copy()
		self.groundold[1] += owner.worldLinearVelocity/60
		self.groundhit = None

	def PS_SetVisible(self):
		owner = self.objects["Root"]
		char = self.objects["Character"]

		if self.camdata["State"] == "FIRST":
			char.setVisible(False, True)
		else:
			char.setVisible(True, True)
		if "SKT" in self.objects:
			for key in self.HAND:
				hand = self.HAND[key]
				skt = self.objects["SKT"].get(hand, None)
				if skt != None:
					skt.setVisible(True, True)

	def PS_Camera(self):
		owner = self.objects["Root"]
		orbit = self.objects["CamOrb"]
		rotate = self.objects["CamRot"]

		up = -self.gravity.normalized()
		if up.length < 0.1:
			up = owner.getAxisVect((0,0,1))
		orbit.alignAxisToVect(up, 2, 1.0)

		xref = orbit.getAxisVect((1,0,0))
		rotate.alignAxisToVect(xref, 0, 1.0)

		self.doCameraCollision()

	## WALKING STATE ##
	def ST_Crouch(self):
		owner = self.objects["Root"]
		char = self.objects["Character"]
		scene = owner.scene

		fac = self.data["CROUCH"]*0.1
		if self.data["CROUCH"] < 0:
			fac = (self.data["CROUCH"]+10)*-0.1

		## EYE HEIGHT ##
		owner.localScale[2] = self.CROUCH_SCALE
		char.localScale[2] = 1/self.CROUCH_SCALE
		char.localPosition[2] = fac*(1/self.CROUCH_SCALE)*self.CROUCH_H

		cr_fac = self.GND_H-(fac*self.CROUCH_H)

		eyevec = [0, 0, (self.EYE_H-self.GND_H)]

		if self.camdata["State"] == "SHOULDER":
			eyevec[0] = self.CAM_SHSIDE
			eyevec[2] -= self.CAM_MIN

		eyevec[2] *= cr_fac

		self.objects["CamRot"].localPosition = eyevec

		## MOVEMENT ##
		ground, angle = self.checkGround()

		owner.setDamping(0, 0)

		if ground != None:
			gpos = ground[1]+(owner.getAxisVect((0,0,1))*(self.gndraybias*cr_fac))
			owner.worldPosition = gpos
			owner.worldLinearVelocity = (0,0,0)

			if ground[0].getPhysicsId() != 0:
				impulse = self.gravity*owner.mass*0.1
				ground[0].applyImpulse(ground[1], impulse, False)

			axis = owner.getAxisVect((0,0,2))
			point = ground[1]+axis
			dist = point-owner.worldPosition

			chkwall = True
			if self.data["CROUCH"] == -20:
				if self.controller.sensors["CROUCH"].status == 3:
					chkwall = False
			elif self.data["CROUCH"] < -5:
				chkwall = False

			if self.checkWall(axis=axis.normalized(), simple=dist.length) != None:
				chkwall = True

		else:
			chkwall = False

		mx = 0
		move = self.motion["Move"].normalized()
		if self.motion["Move"].length > 0.01:
			mx = 0.02

		self.doMovement((move[0], move[1], 0), mx)
		self.doMoveAlign()

		if chkwall == False and self.data["CROUCH"] < -5:
			self.doPlayerAnim("IDLE", blend=10)

			if self.data["CROUCH"] == -10:
				self.data["CROUCH"] = 0
				self.doCrouch(False)
			else:
				self.data["CROUCH"] += 1

		else:
			self.doPlayerAnim("CROUCH", blend=10)

			if self.data["CROUCH"] == 10:
				self.data["CROUCH"] = -20
			elif self.data["CROUCH"] >= 0:
				self.data["CROUCH"] += 1

	def ST_Walking(self):
		owner = self.objects["Root"]
		char = self.objects["Character"]
		scene = owner.scene

		move = self.motion["Move"].normalized()

		ground, angle = self.checkGround()

		owner.setDamping(0, 0)

		if ground != None:
			gpos = ground[1]+(owner.getAxisVect((0,0,1))*self.gndraybias)

			if self.jump_state in ["FALLING", "A_JUMP", "B_JUMP", "NO_AIR"] and owner.localLinearVelocity[2] < 0.1:
				self.jump_timer = 0
				self.jump_state = "NONE"
				if ground[0].getPhysicsId() != 0:
					impulse = (owner.worldLinearVelocity+self.gravity)*owner.mass*0.1
					ground[0].applyImpulse(ground[1], impulse, False)
				#Z = owner.localLinearVelocity[2]
				#if Z < -8:
				#	self.data["HEALTH"] += (Z+8)*self.GRAV_DAMAGE
				self.doAnim(STOP=True)
				self.resetAcceleration(owner.worldLinearVelocity*(1/60))
				self.accel_timer = 10

			if self.jump_state == "NONE":
				owner.worldPosition = gpos
				owner.worldLinearVelocity = (0,0,0)

				self.doPlayerOnGround()

				if ground[0].getPhysicsId() != 0:
					impulse = self.gravity*owner.mass*0.1
					ground[0].applyImpulse(ground[1], impulse, False)

			elif self.jump_state in ["JUMP", "JP_WAIT"]:
				self.jump_state = "JP_WAIT"
				self.jump_timer += 1

				if self.jump_timer > 10:
					self.doAnim(STOP=True)
					self.jump_timer = 0
					self.jump_state = "NONE"
			else:
				gdif = owner.worldOrientation.inverted()*(owner.worldPosition-gpos)
				if gdif[2] < 0:
					owner.worldPosition = gpos
					owner.localLinearVelocity[2] = 0

				if self.controller.sensors["JUMP"].status == 2:
					self.jump_state = "B_JUMP"
				else:
					self.doAnim(STOP=True)
					self.jump_state = "FALLING"

		else:
			self.doPlayerAnim("FALLING")

			if self.jump_state in ["FALLING", "A_JUMP", "B_JUMP", "NO_AIR"]:
				if self.camdata["State"] != "THIRD":
					self.doMoveAlign(up=False)

				if self.gravity.length < 0.1:
					vref = (self.motion["Move"][0], self.motion["Move"][1], self.motion["Climb"])
					vref = self.objects["CamOrb"].getAxisVect(vref)
					vref.normalize()
					owner.applyForce(vref*1, False)

				if self.jump_state != "NO_AIR" and self.gravity.length >= 0.1:
					vref = self.objects["CamOrb"].getAxisVect([move[0], move[1], 0])
					owner.applyForce(vref*5, False)

			if self.jump_state in ["NONE", "JUMP"]:
				if self.jump_state == "NONE" and self.gravity.length >= 0.1:
					if self.controller.sensors["JUMP"].status == 2 and self.jump_timer >= 1:
						self.doJump(move=0.8, align=True)
					elif self.controller.sensors["CROUCH"].status == 0:
						self.doJump(height=1, move=0.5)
				self.jump_state = "A_JUMP"

			if self.controller.sensors["JUMP"].status == 2:
				if self.jump_state == "FALLING":
					self.jump_state = "B_JUMP"
			else:
				self.jump_state = "FALLING"

			self.jump_timer += 1

		self.alignToGravity()

	## RUN ##
	def RUN(self):
		for run in self.active_pre:
			st = run()
		self.getInputs()
		self.active_state()
		for run in self.active_post:
			st = run()




